"""Station objects for veliberator"""
from xml.dom.minidom import parse

from veliberator.models import db_connection
from veliberator.models import StationInformation
from veliberator.status import StationStatus

from veliberator.geofinder import StationGeoFinder

class UnknowStation(Exception):
    pass

class Station(object):
    """Station object,
    combinating informations and status"""

    def __init__(self, velib_id):
        self.id = int(velib_id)
        self.informations = self.acquire_informations()

        if not self.informations:
            raise UnknowStation('The Station ID does not exist.')

        self.status = StationStatus(self.id)
        self.finder = StationGeoFinder(self)

    def acquire_informations(self):
        """Default method for acquiring informations
        of a stations, by connecting to the database"""
        try:
            return StationInformation.get(self.id)
        except AttributeError:
            db_connection()
            return StationInformation.get(self.id)

    @property
    def is_open(self):
        """Property telling if the station is open"""
        return self.informations.opened

    @property
    def is_bonus(self):
        """Property telling if the station is bonus"""
        return self.informations.bonus

    def is_free(self, places=1):
        """Method property if the station
        has free places"""
        return self.is_open and self.status.free >= places

    def is_available(self, places=1):
        """Method property if the station
        has available places"""
        return self.is_open and self.status.available >= places

    @property
    def stations_around(self):
        """Find the stations around"""
        return self.finder.get_stations_around()

    def __repr__(self):
        if self.informations:
            return '<Station "%s" (%s)>' % (self.id, self.informations.address)
        return '<Station "%s">' % self.id






"""Unit tests for GeoFinder objects"""
import unittest

from veliberator.geofinder import global_geofinder_cache
from veliberator.geofinder import BaseGeoFinder
from veliberator.geofinder import StationGeoFinder
from veliberator.geofinder import AddressGeoFinder
from veliberator.geofinder import GeoFinderError
from veliberator.models import StationInformation
from veliberator import Cartography
from veliberator import Station
from veliberator.settings import TEST_XML_URL_DATA_STATION

class BaseGeoFinderTestCase(unittest.TestCase):

    def test_ComputeSquareArea(self):
        finder = BaseGeoFinder()
        finder.lat = 1
        finder.lng = 1
        self.assertEquals(finder.compute_square_area(),
                          (('0.99', '1.00', '1.01'), ('0.99', '1.00', '1.01')))
        finder.lat = '1'
        finder.lng = '1'
        self.assertEquals(finder.compute_square_area(),
                          (('0.99', '1.00', '1.01'), ('0.99', '1.00', '1.01')))
        finder.lat = 4.897645
        finder.lng = 18.798923
        self.assertEquals(finder.compute_square_area(),
                          (('4.89', '4.90', '4.91'), ('18.79', '18.80', '18.81')))

    def test_ComputeStationDistances(self):
        finder = BaseGeoFinder()
        finder.lat = 1
        finder.lng = 1

        si1 = StationInformation(id=42, lat=1, lng=1)
        result = finder.compute_station_distances([si1])
        self.assertEquals(result, {})

        si2 = StationInformation(id=43, lat=4, lng=5)
        result = finder.compute_station_distances([si2,])
        self.assertEquals(result, {si2: 5.0})

        si3 = StationInformation(id=44, lat=8, lng=10)
        si4 = StationInformation(id=45, lat=6, lng=5.5)
        result = finder.compute_station_distances([si1, si2, si3, si4])
        self.assertEquals(len(result), 3)

        StationInformation.query.delete()

    def test_GetStationsInArea(self):
        test_1 = (('4.89', '4.90', '4.91'), ('18.79', '18.80', '18.81'))
        test_2 = (('48.80', '48.81', '48.82'), ('2.37', '2.38', '2.39'))

        Cartography.synchronize(TEST_XML_URL_DATA_STATION)
        finder = BaseGeoFinder()

        self.assertEquals(len(finder.get_stations_in_area(*test_1)), 0)
        self.assertEquals(len(finder.get_stations_in_area(*test_2)), 16)
                
        Cartography.flush()

    def test_GetStationsAround(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)

        finder = BaseGeoFinder()
        finder.lat = 1
        finder.lng = 1
        self.assertEquals(len(finder.get_stations_around()), 0)

        finder.lat = 48.81
        finder.lng = 2.38
        self.assertEquals(finder.get_stations_around(),                     
                          [42012, 42010, 42009, 42008, 42016, 42006, 42007, 42015,
                           42014, 42002, 42001, 42011, 42003, 13055, 42004, 42201])
        
        Cartography.flush()

    def test_Cache(self):
        key = (1, 1)
        self.assertFalse(global_geofinder_cache.has_key(key))        
        finder = BaseGeoFinder()
        finder.lat = key[0]
        finder.lng = key[1]
        finder.get_stations_around()
        self.assertTrue(global_geofinder_cache.has_key(key))


class StationGeoFinderTestCase(unittest.TestCase):

    def test_Init(self):
        Cartography.synchronize(TEST_XML_URL_DATA_STATION)

        station = Station(42008)
        geofinder = StationGeoFinder(station)
        self.assertEquals(geofinder.lat, station.informations.lat)
        self.assertEquals(geofinder.lng, station.informations.lng)

        Cartography.flush()

class AddressGeoFinderTestCase(unittest.TestCase):

    def test_Init(self):
        address = '1 place de la Bastille, 75012 Paris'
        geofinder = AddressGeoFinder(address)
        self.assertEquals(geofinder.precision, 8)
        self.assertEquals(geofinder.lat, '48.8530893')
        self.assertEquals(geofinder.lng, '2.3687316')

        address = 'Bethune, 62400'
        self.assertRaises(GeoFinderError, AddressGeoFinder, address)

    def test_GeoCompute(self):
        address = '1 place de la Bastille, 75012 Paris'
        geofinder = AddressGeoFinder(address)
        self.assertEquals(len(geofinder.geocompute(address)), 3)

suite = unittest.TestSuite([
    unittest.TestLoader().loadTestsFromTestCase(BaseGeoFinderTestCase),
    unittest.TestLoader().loadTestsFromTestCase(StationGeoFinderTestCase),
    unittest.TestLoader().loadTestsFromTestCase(AddressGeoFinderTestCase),])




"""Objects for helping filling data"""
from urllib import urlopen
from xml.dom.minidom import parse

from veliberator.settings import XML_URL_DATA_STATION
from veliberator.xml_wrappers import xml_station_information_wrapper
from veliberator.models import StationInformation, session


class Cartography(object):
    """Grab the data and save it in db"""

    @staticmethod
    def synchronize(url=XML_URL_DATA_STATION):
        dom = parse(urlopen(url))
        for marker in dom.getElementsByTagName('marker'):
            values = xml_station_information_wrapper(marker)
            station = StationInformation.update_or_create(values, surrogate=False)

        session.commit()

    @staticmethod
    def flush():
        StationInformation.query.delete()




"""Geo localizator objects"""
from math import sqrt
from urllib import quote_plus
from urllib import urlopen

from veliberator.models import StationInformation

global_geofinder_cache = {}

def cache_wrapper(method):
    def cache(instance):
        """Use a global cache for the results
        of stations around"""
        global global_geofinder_cache
        
        key_cache = (instance.lat, instance.lng)
        if not key_cache in global_geofinder_cache.keys():
            global_geofinder_cache[key_cache] = method(instance)
       
        return global_geofinder_cache[key_cache]
    return cache

class GeoFinderError(Exception):
    pass

class BaseGeoFinder(object):
    """Base GeoFinder object,
    for finding stations around"""
    lat = None
    lng = None

    def compute_square_area(self):
        """Round the GPS coordonates to a wide area"""
        #to be refactored for configuring the area size
        lat_orig = float(self.lat)
        lat_pos = '%.2f' % (lat_orig + 0.01)
        lat_neg = '%.2f' % (lat_orig - 0.01)
        lat_orig = '%.2f' % lat_orig

        lng_orig = float(self.lng)
        lng_pos = '%.2f' % (lng_orig + 0.01)
        lng_neg = '%.2f' % (lng_orig - 0.01)
        lng_orig = '%.2f' % lng_orig

        return (lat_neg, lat_orig, lat_pos), (lng_neg, lng_orig, lng_pos)

    def compute_station_distances(self, stations):
        """Compute the distance from the coords,
        to the other stations, Pythagore powered"""
        distances = {}

        for station in stations:
            if station.lat == self.lat and station.lng == self.lng:
                continue
            l1 = float(self.lat) - float(station.lat)
            l2 = float(self.lng) - float(station.lng)
            dist = sqrt(pow(l1, 2) + pow(l2, 2))
            distances[station] = dist

        return distances

    def get_stations_in_area(self, lats, lngs):
        """Make a query to find the station with
        lat and lng matching with the area"""
        return StationInformation.query.filter(
            (StationInformation.lat.startswith(lats[0]) |
             StationInformation.lat.startswith(lats[1]) |
             StationInformation.lat.startswith(lats[2])) &
            (StationInformation.lng.startswith(lngs[0]) |
             StationInformation.lng.startswith(lngs[1]) |
             StationInformation.lng.startswith(lngs[2]))).all()


    @cache_wrapper
    def get_stations_around(self):
        """Find the stations around the lat and lng,
        sorted by proximity"""
        lats, lngs = self.compute_square_area()
        stations = self.get_stations_in_area(lats, lngs)
        station_distances = self.compute_station_distances(stations)
        station_distances_ordered = sorted(station_distances.iteritems(),
                                           key=lambda (k,v): (v,k))
        stations_id = [station.id for station, distance in station_distances_ordered]

        return stations_id


class StationGeoFinder(BaseGeoFinder):

    def __init__(self, station):
        """Init the object with a Station object"""
        self.lat = station.informations.lat
        self.lng = station.informations.lng


class AddressGeoFinder(BaseGeoFinder):

    def __init__(self, address):
        """Init the object with an address
        who will be geocomputed"""
        self.address = address
        precision, lat, lng = self.geocompute(address)

        self.lat = lat
        self.lng = lng
        self.precision = int(precision)

        if self.precision <= 6:
            raise GeoFinderError('Your address is not available.')

    def geocompute(self, address):
        """Geocompute an address with GMap"""
        address = quote_plus(address)
        request = "http://maps.google.com/maps/geo?sensor=false&q=%s&output=%s&oe=utf8&gl=fr" % (
            address, 'csv')
        data = urlopen(request).read().split(',')
        if data[0] != '200':
            raise GeoFinderError('Service unavailable')
        return data[1:]


